/*
 * Copyright (c) 2006 Joris Vink <joris@openbsd.org>
 *
 * Permission to use, copy, modify, and distribute this software for any
 * purpose with or without fee is hereby granted, provided that the above
 * copyright notice and this permission notice appear in all copies.
 *
 * THE SOFTWARE IS PROVIDED "AS IS" AND THE AUTHOR DISCLAIMS ALL WARRANTIES
 * WITH REGARD TO THIS SOFTWARE INCLUDING ALL IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS. IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR
 * ANY SPECIAL, DIRECT, INDIRECT, OR CONSEQUENTIAL DAMAGES OR ANY DAMAGES
 * WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN AN
 * ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
 * OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE.
 */

#ifndef REPO_H
#define REPO_H

#define CVS_LOCK		"#cvs.lock"
#define CVS_JUNK		"#cvs."
#define CVS_LOCK_SLEEP		30
#define CVS_LOCK_TRIES		5

#define REPOSITORY_DODIRS	0x01
#define REPOSITORY_IS_ATTIC	0x02

extern struct wklhead repo_locks;

void	cvs_repository_unlock(const char *);
void	cvs_repository_lock(const char *, int);
void	cvs_repository_getdir(const char *, const char *,
		struct cvs_flisthead *, struct cvs_flisthead *, int);

#endif
